"""
Generate a random trecvid video

Guidelines: "Building video queries for Trecvid2008 copy detection task"

"""



import random,types,math


from graphic_ops import *


#################################################################
# Sum up availabe input files


# Trecvid input clips. We need an auxiliary file trecvidlist.txt that
# gives the number of frames of each video.
trecviddir="/scratch/clear/dataset/video/trecvid/mpg/all/"
trecvidfiles=[l.split() for l in open("trecvidlist.txt","r")]
trecvidfiles=[(trecviddir+fname,int(nframe_string)) for (fname,nframe_string) in trecvidfiles]


# distractors are assumed to be of arbitrary length. We take them from
# DVD .VOB files (from 6 seasons of "Buffy the Vampire Slayer")
vobdir="/home/clear/marszale/data/buffy/videos/"
vobfiles=[vobdir+fname for fname in os.listdir(vobdir) if fname.endswith(".vob")]


# occluding images are semi-transparent PNG's of 352 by 288 pixels
maskdir="./video_masks/"
masks=[maskdir+fname for fname in os.listdir(maskdir) if fname.endswith(".png")]

print "Found %d trecvids, %d buffies, %d masks"%(len(trecvidfiles),len(vobfiles),len(masks))

# hardcoded for all videos
fps=25.0

# output filename will be outputpattern%(step1no,step2no)
outputpattern="/home/clear/douze/sample3/val_%03d_%03d.mpg"


#################################################################
# Step1



def obj_to_dict(self):
  d={"classname":self.__class__.__name__}
  for attrname in dir(self):
    if attrname.startswith("__"): continue
    val=getattr(self,attrname)
    if type(val) in [types.IntType,types.FloatType,types.StringType,
                     types.ListType,types.TupleType]:
      d[attrname]=val
  return d


class Step1:
  """ first step: combination of a ref and a nonref video"""

  def __init__(self):
    
    # computations in sec
    while True:
      # pick a random input video
      (ref_name,ref_len)=random.choice(trecvidfiles)
      ref_len=ref_len/fps

      # don't use a 60 s margin at the beginning and end because of
      # into/exit material
      if ref_len<=120: continue
      tmin=60
      tmax=ref_len-60

      # random beginning time
      tcref=random.random()*(tmax-tmin)+tmin

      if not (tcref+3<tmax): continue

      # random length (between 3 s and 1 min)
      lref=3+random.random()*(min(tmax-tcref,60)-3)

      # ok we found something suitable
      break

    self.ref_name,self.tcref,self.lref=ref_name,tcref,lref
    
    nonref_name=random.choice(vobfiles)

    # assume Buffy is as long as we need
    # skip 6 min at beginning to avoid credits list in the beginning
    tcnonref=240.0+400.0*random.random()    
    lnonref_min=max(30,lref)
    lnonref=lnonref_min+random.random()*(3*60-lnonref_min)

    self.nonref_name,self.tcnonref,self.lnonref=nonref_name,tcnonref,lnonref  
    
    # how should the videos be combined?
    
    self.random_type=1+random.randrange(3)

    # random_type==1: we only keep the reference video    
    if self.random_type==2:
      # insert ref video at random location in nonref video
      self.t_insert=(lnonref-lref)*random.random()
    elif self.random_type==3:
      # don't use ref video.
      self.t_insert=self.lref=0.0

  def make_Stage1Video(self):
    lref_f=int(self.lref*fps)
    if self.random_type==1:
      return Stage1Video(self.ref_name,0,lref_f,lref_f,self.tcref)
    if self.random_type in (2,3):
      t_insert_f=int(self.t_insert*fps)
      return Stage1Video(self.ref_name,
                         t_insert_f,t_insert_f+lref_f,int(self.lnonref*fps),
                         self.tcref,
                         self.nonref_name,self.tcnonref)
    
  def __str__(self):
    return "Step1(%s)"%obj_to_dict(self)




#################################################################
# Step2





def rand_motion(d,strength):
  if d==0:
    return random.random()*strength
  else:
    return d-random.random()*strength

def random_log_uniform(vmin,vmax):
  vmin=math.log(vmin)
  vmax=math.log(vmax)  
  return math.exp(random.random()*(vmax-vmin)+vmin)



class SetBitrate:
  " dummy filter "
  def __init__(self):
    self.bitrate=int(random_log_uniform(6,100))

def make_random_vignette(xd,yd):
  bgvid_t=240+600.0*random.random()
  bgvid=VOBPlayer(random.choice(vobfiles),bgvid_t,xd,yd)
  scale=0.3+0.2*random.random()
  dx=xd*(1-scale)*random.random()
  dy=yd*(1-scale)*random.random()     
  return PictureInVid(bgvid,scale,dx,dy)    
  
class Step2Single:
  " A single step2 "
  def __init__(self,xd,yd):
    r=random.randrange(5)
    self.random_type=r
    if r==0:
      # 4*2 coordinates of perspective distoriton
      coords=[rand_motion(d,40) for y in (0,yd) for x in (0,xd) for d in (x,y)]
      self.filter=CamcordingFilter(coords)
    elif r==1:
      self.filter=make_random_vignette(xd,yd)
    elif r==2:
      self.filter=OverlayImageFilter(random.choice(masks))
    elif r==3:
      # has to be applied on output      
      self.filter=SetBitrate()
    elif r==4:
      self.filter=GammaFilter(random_log_uniform(0.4,2.5))

  def __str__(self):
    return "Step2Single(%s)"%obj_to_dict(self.filter)



class Step2Multiple:
  
  def __init__(self,xd,yd):
    " not exactly as described in the reference document " 
    self.filters=[]
    for i in range(random.randrange(3)):
      self.filters.append(self.decrease_qual(xd,yd))
    for i in range(random.randrange(3)):
      self.filters.append(self.post_prod(xd,yd))

  def decrease_qual(self,xd,yd):    
    r=random.randrange(6)
    if r==0:
      return BlurFilter(random.randrange(1,7))
    if r==1:
      return FrameDropper(0.07+0.2*random.random())
    if r==2:
      return AspectRatioFilter(0.8+0.4*random.random())
    if r==3:
      return NoiseFilter(0.1+0.4*random.random())
    if r==4:
      return ContrastFilter(random.randrange(-80,80))
    if r==5:
      return PixelizeFilter(random.randrange(4,9))

  def post_prod(self,xd,yd):
    r=random.randrange(6)
    if r==0:
      return BlackMargins([random.randrange(40) for i in range(4)])
    if r==1:
      return BlackBorderFilter((random.randrange(-80,80),random.randrange(-80,80)))
    if r==2:
      return TScaler(0.8+0.4*random.random())
    if r==3:
      return FlipFilter()
    if r==4:
      return make_random_vignette(xd,yd)
    if r==5:
      # handle on output
      return SetBitrate()
    
  def __str__(self):
    return "Step2Multiple(%s)"%[obj_to_dict(filt) for filt in self.filters]



#################################################################
# Main function
      

                     
                     
def generate(step1no,step2no,dryrun=False):

  try:

    random.seed(step1no)      
    
    step1=Step1()

    print "Seeds: %d %d"%(step1no,step2no)

    print "step1=",step1
    vid=step1.make_Stage1Video()

    random.seed(step2no)
    
    if random.random()<0.3:
      step2=Step2Single(vid.w,vid.h)
    else:
      step2=Step2Multiple(vid.w,vid.h)   

    print "step2=",step2   

    if dryrun: return

    bitrate=512

    if isinstance(step2,Step2Single):
      filters=[step2.filter]
    else:
      filters=step2.filters

    # filters may contain pre-filters and bitrate setting
    filters2=[]
    for filt in filters:
      if isinstance(filt,SetBitrate):
        bitrate=filt.bitrate
      elif isinstance(filt,FilterVidPlayer):
        # insert at beginning of chain
        filt.set_vid(vid)
        vid=filt
      else:
        filters2.append(filt)

    sink=VidSink(outputpattern%(step1no,step2no),
                 vid.w,vid.h,bitrate=bitrate)     

    print "running filters"   
    
    run_filter(vid,filters2,sink)
    
  except Exception,e:
    # need to take care of exceptions here as the GIMP plugin system
    # tends to hide stacktraces
    traceback.print_exc(50,sys.stdout)
    pdb.gimp_quit(0)
    sys.exit(1)




